/* BOGL - Ben's Own Graphics Library.
   Written by Ben Pfaff <pfaffben@debian.org>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA. */

/*#define NDEBUG*/
#include <assert.h>
#include <string.h>
#include "bogl.h"
#include "boglP.h"
#include "bogl-cfb8.h"

/* Set pixel (X,Y) to color C. */
void
bogl_cfb8_pixel (int x, int y, int c)
{
  bogl_drawing = 1;

  assert (x >= 0 && x < bogl_xres);
  assert (y >= 0 && y < bogl_yres);
  assert (c >= 0 && c < bogl_ncols);
  
  bogl_frame[x + y * bogl_line_len] = c;

  bogl_drawing = 0;
}

/* Paint a horizontal line from (X1,Y) to (X2,Y) in color C, where X2
   >= X1.  The final point is not painted. */
void
bogl_cfb8_hline (int x1, int x2, int y, int c)
{
  assert (x1 >= 0 && x1 < bogl_xres);
  assert (x2 >= 0 && x2 <= bogl_xres);
  assert (x2 >= x1);
  assert (y >= 0 && y < bogl_yres);
  assert (c >= 0 && c < bogl_ncols);

  if (x1 == x2)
    return;

  bogl_drawing = 1;
  memset ((char *) bogl_frame + x1 + y * bogl_line_len, c, x2 - x1);
  bogl_drawing = 0;
}

/* Paints a vertical line from (X,Y1) to (X,Y2) in color C.  The final
   point is not painted. */
void
bogl_cfb8_vline (int x, int y1, int y2, int c)
{
  assert (x >= 0 && x < bogl_xres);
  assert (y1 >= 0 && y1 < bogl_yres);
  assert (y2 >= 0 && y2 <= bogl_yres);
  assert (y2 >= y1);
  assert (c >= 0 && c < bogl_ncols);

  bogl_drawing = 1;
  for (; y1 < y2; y1++)
    bogl_frame[x + y1 * bogl_line_len] = c;
  bogl_drawing = 0;
}

/* Clear the region from (X1,Y1) to (X2,Y2) to color C, not including
   the last row or column.  If C == -1 then the region's colors are
   inverted rather than set to a particular color.  */
void
bogl_cfb8_clear (int x1, int y1, int x2, int y2, int c)
{
  volatile char *dst;
  
  assert (x1 >= 0 && x1 < bogl_xres);
  assert (x2 >= 0 && x2 <= bogl_xres);
  assert (x2 >= x1);
  assert (y1 >= 0 && y1 < bogl_yres);
  assert (y2 >= 0 && y2 <= bogl_yres);
  assert (y2 >= y1);
  assert (c >= -1 && c < bogl_ncols);

  if (x1 == x2)
    return;

  bogl_drawing = 1;
  dst = bogl_frame + x1 + y1 * bogl_line_len;
  for (; y1 < y2; y1++)
    {
      memset ((char *) dst, c, x2 - x1);
      dst += bogl_line_len;
    }
  bogl_drawing = 0;
}

void
bogl_cfb8_text (int xx, int yy, const char *s, int n, int fg, int bg,
		struct bogl_font *font)
{
  int h;
  
  assert (xx >= 0 && xx < bogl_xres);
  assert (yy >= 0 && yy < bogl_yres);
  assert (fg >= 0 && fg < bogl_ncols);
  assert (bg >= -1 && bg < bogl_ncols);

  bogl_drawing = 1;

  h = font->height;
  if (yy + h > bogl_yres)
    h = bogl_yres - yy;

  for (; n--; s++)
    {
      volatile char *dst = bogl_frame + xx + yy * bogl_line_len;

      const unsigned char ch = *s;
      const unsigned long *character = &font->content[font->offset[ch]];
      int w = font->width[ch];

      int x, y;

      if (xx + w > bogl_xres)
	w = bogl_xres - xx;
      
      for (y = 0; y < h; y++)
	{
	  unsigned long c = *character++;
	  
	  for (x = 0; x < w; x++)
	    {
	      if (c & 0x80000000)
		dst[x] = fg;
	      else if (bg != -1)
		dst[x] = bg;

	      c <<= 1;
	    }

	  dst += bogl_line_len;
	}

      xx += w;
      if (xx >= bogl_xres)
	break;
    }

  bogl_drawing = 0;
}

/* Write PIXMAP at location (XX,YY), with the pixmap's colors mapped
   according to COLOR_MAP. */
void
bogl_cfb8_put (int xx, int yy, const struct bogl_pixmap *pixmap,
	       const int color_map[16])
{
  volatile char *dst;
  const unsigned char *src;
  int h;
  
  assert (xx + pixmap->width <= bogl_xres);
  assert (yy >= 0 && yy < bogl_yres);
  assert (yy + pixmap->width <= bogl_yres);
  src = pixmap->data;

  bogl_drawing = 1;

  h = pixmap->height;
  dst = bogl_frame + xx + yy * bogl_line_len;
  while (h--)
    {
      int w = pixmap->width;
      while (w)
	{
	  int color = *src & 0xf;
	  int count = *src >> 4;
	  src++;

	  w -= count;
	  while (count--)
	    *dst++ = color_map[color];
	}

      dst += bogl_line_len - pixmap->width;
    }

  bogl_drawing = 0;
}

/* Initialize CFB8 mode.  Returns the number of bytes to mmap for the
   framebuffer. */
size_t
bogl_cfb8_init (void)
{
  return bogl_xres * bogl_yres;
}
