/* Copyright (c) 1991 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Generic device dependent routines.
 * All of these routines use other device routines to perform their work,
 * eventually getting to real device routines.
 */
 
#define _GNU_SOURCE 1
#include <assert.h>
#include <fcntl.h>
#include <limits.h>
#include <linux/fb.h>
#include <linux/kd.h>
#include <linux/vt.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <termios.h>
#include <sys/io.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include "bogl.h"
#include "boglP.h"
#include "bogl-vga16.h"
#include "bogl-cfb8.h"


/*===========================================================================*
 *				bogl_drawline				     *
 *===========================================================================*/
void bogl_line(int x1, int y1, int x2, int y2, int color)
{
/* Generic draw line routine. */
  int temp;
  int xdelta;
  int ydelta;
  int xinc;
  int yinc;
  int rem;

  if (y1 == y2) {
	bogl_hline(x1, x2, y1, color);
	return;
  }
  if (x1 == x2) {
	bogl_vline(x1, y1, y2, color);
	return;
  }
  xdelta = x2 - x1;
  ydelta = y2 - y1;
  if (xdelta < 0) xdelta = -xdelta;
  if (ydelta < 0) ydelta = -ydelta;
  xinc = (x2 > x1) ? 1 : -1;
  yinc = (y2 > y1) ? 1 : -1;
  bogl_pixel(x1, y1, color);
  if (xdelta >= ydelta) {
	rem = xdelta / 2;
	do {
		x1 += xinc;
		rem += ydelta;
		if (rem >= xdelta) {
			rem -= xdelta;
			y1 += yinc;
		}
		bogl_pixel(x1, y1, color);
	} while (x1 != x2);
  } else {
	rem = ydelta / 2;
	do {
		y1 += yinc;
		rem += xdelta;
		if (rem >= ydelta) {
			rem -= ydelta;
			x1 += xinc;
		}
		bogl_pixel(x1, y1, color);
	} while (y1 != y2);
  }
}


/*===========================================================================*
 *				draw4points				     *
 *===========================================================================*/
static void draw4points(int x, int y,int px,int  py,int  color)
{
/* Set four points symmetrically situated around a point. */
  bogl_pixel (x + px, y + py, color);
  bogl_pixel (x - px, y + py, color);
  bogl_pixel (x + px, y - py, color);
  bogl_pixel (x - px, y - py, color);
}


/*===========================================================================*
 *				bogl_drawellipse				     *
 *===========================================================================*/
void bogl_drawellipse(x, y, rx, ry, color)
int x, y;			/* coordinates of center of ellipse */
int rx;			/* radius along x axis */
int ry;			/* radius along y axis */
{
/* Generic ellipse drawing routine. */
  int xp, yp;			/* current point (based on center) */
  long Asquared;		/* square of x semi axis */
  long TwoAsquared;
  long Bsquared;		/* square of y semi axis */
  long TwoBsquared;
  long d;
  long dx, dy;

  if ((rx < 0) || (ry < 0)) return;

  xp = 0;
  yp = ry;
  Asquared = rx * rx;
  TwoAsquared = 2 * Asquared;
  Bsquared = ry * ry;
  TwoBsquared = 2 * Bsquared;
  d = Bsquared - Asquared * ry + (Asquared >> 2);
  dx = 0;
  dy = TwoAsquared * ry;

  while (dx < dy) {
	draw4points(x, y, xp, yp, color);
	if (d > 0) {
		yp--;
		dy -= TwoAsquared;
		d -= dy;
	}
	xp++;
	dx += TwoBsquared;
	d += (Bsquared + dx);
  }
  d += ((3L * (Asquared - Bsquared) / 2L - (dx + dy)) >> 1);
  while (yp >= 0) {
	draw4points(x, y, xp, yp, color);
	if (d < 0) {
		xp++;
		dx += TwoBsquared;
		d += dx;
	}
	yp--;
	dy -= TwoAsquared;
	d += (Asquared - dy);
  }
}



/*===========================================================================*
 *				bogl_fillellipse				     *
 *===========================================================================*/
void bogl_fillellipse(int x, int y, int rx, int ry, int color)
{
/* Generic fill ellipse routine. */
  int xp, yp;			/* current point (based on center) */
  long Asquared;		/* square of x semi axis */
  long TwoAsquared;
  long Bsquared;		/* square of y semi axis */
  long TwoBsquared;
  long d;
  long dx, dy;

  if ((rx < 0) || (ry < 0)) return;

  xp = 0;
  yp = ry;
  Asquared = rx * rx;
  TwoAsquared = 2 * Asquared;
  Bsquared = ry * ry;
  TwoBsquared = 2 * Bsquared;
  d = Bsquared - Asquared * ry + (Asquared >> 2);
  dx = 0;
  dy = TwoAsquared * ry;

  while (dx < dy) {
	bogl_hline(x - xp, x + xp, y - yp, color);
	bogl_hline(x - xp, x + xp, y + yp, color);
	if (d > 0) {
		yp--;
		dy -= TwoAsquared;
		d -= dy;
	}
	xp++;
	dx += TwoBsquared;
	d += (Bsquared + dx);
  }
  d += ((3L * (Asquared - Bsquared) / 2L - (dx + dy)) >> 1);
  while (yp >= 0) {
	bogl_hline(x - xp, x + xp, y - yp, color);
	bogl_hline(x - xp, x + xp, y + yp, color);
	if (d < 0) {
		xp++;
		dx += TwoBsquared;
		d += dx;
	}
	yp--;
	dy -= TwoAsquared;
	d += (Asquared - dy);
  }
}


