/* Copyright (c) 1991 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routines to determine clipping regions.
 */

#include "bogl.h"

/* Clip cache rectangle information.
 * After calling clippoint, this rectangle is guaranteed to contain the
 * specified point (among others), and all points in the rectangle are
 * plottable or not according to the value of bogl_clipresult.
 */
 
int bogl_clipminx;		/* minimum x value of cache rectangle */
int bogl_clipminy;		/* minimum y value of cache rectangle */
int bogl_clipmaxx;		/* maximum x value of cache rectangle */
int bogl_clipmaxy;		/* maximum y value of cache rectangle */
int bogl_clipresult;		/* whether clip rectangle is plottable */

static int clipcount;	/* number of clip rectangles */
static struct bogl_cliprect cliprects[GR_MAX_CLIPRECTS];	/* clip rectangles */


/*===========================================================================*
 *				clipinit				     *
 *===========================================================================*/
void bogl_clipinit(int count, struct bogl_cliprect *table)
{
/* Specify a new set of clip rectangles.
 * This should only be called after the device has been initialized.
 * If zero clip rectangles are specified, then clipping is done to the
 * full device area.  All clip rectangles are modified if necessary to
 * lie within the device area.
 */
  struct bogl_cliprect *rp;		/* current rectangle */

  /* If there are no clip rectangles, then default to the full device area. */
  if (count <= 0) {
	bogl_clipminx = 0;
	bogl_clipminy = 0;
	bogl_clipmaxx = bogl_xres - 1;
	bogl_clipmaxy = bogl_yres - 1;
	clipcount = 0;
	bogl_clipresult = 1;
	return;
  }

  /* Copy the clip table to our own static array, modifying each
   * rectangle as necesary to fit within the device area.  If the clip
   * rectangle lies entirely outside of the device area, then skip it.
   */
  rp = cliprects;
  clipcount = 0;
  if (count > GR_MAX_CLIPRECTS) count = GR_MAX_CLIPRECTS;
  while (count-- > 0) {
	*rp = *table++;
	if (rp->x < 0) {
		rp->width += rp->x;
		rp->x = 0;
	}
	if (rp->y < 0) {
		rp->height += rp->y;
		rp->y = 0;
	}
	if ((rp->x >= bogl_xres) || (rp->width <= 0) ||
	    (rp->y >= bogl_yres) || (rp->height <= 0))
		continue;
	if (rp->x + rp->width > bogl_xres) rp->width = bogl_xres - rp->x;
	if (rp->y + rp->height > bogl_yres)
		rp->height = bogl_yres - rp->y;
	rp++;
	clipcount++;
  }

  /* If there were no surviving clip rectangles, then set the clip
   * cache to prevent all drawing.
   */
  if (clipcount == 0) {
	bogl_clipminx = 0;
	bogl_clipminy = 0;
	bogl_clipmaxx = 65535;
	bogl_clipmaxy = 65535;
	bogl_clipresult = 0;
	return;
  }

  /* There was at least one valid clip rectangle. Default the clip
   * cache to be the first clip rectangle.
   */
  bogl_clipminx = cliprects[0].x;
  bogl_clipminy = cliprects[0].y;
  bogl_clipmaxx = bogl_clipminx + cliprects[0].width - 1;
  bogl_clipmaxy = bogl_clipminy + cliprects[0].height - 1;
  bogl_clipresult = 1;
}


/*===========================================================================*
 *				clippoint				     *
 *===========================================================================*/
int bogl_clippoint(int x,int y)
{
/* Check a point against the list of clip rectangles.
 * Returns TRUE if the point is within one or more rectangles and thus
 * can be plotted, or FALSE if the point is not within any rectangle and
 * thus cannot be plotted.  Also remembers the coordinates of a clip cache
 * rectangle containing the specified point such that every point in the
 * rectangle would give the same result.  By examining this clip cache
 * rectangle after a call to this routine, the caller can efficiently
 * check many nearby points without needing any further calls.
 */
  int count;
  struct bogl_cliprect *rp;
  int temp;

  /* First see whether the point lies within the current clip cache
   * rectangle.  If so, then we already know the result.
   */
  if ((x >= bogl_clipminx) && (x <= bogl_clipmaxx) &&
      (y >= bogl_clipminy) && (y <= bogl_clipmaxy)) {
	return bogl_clipresult;
  }

  /* If the point is outside of the screen area, then it is not
   * plottable, and the clip cache rectangle is the whole half-plane
   * outside of the screen area.
   */
  if (x < 0) {
	bogl_clipminx = 0;
	bogl_clipmaxx = -1;
	bogl_clipminy = 0;
	bogl_clipmaxy = 65535;
	bogl_clipresult = 0;
	return 0;
  }
  if (y < 0) {
	bogl_clipminx = 0;
	bogl_clipmaxx = 65535;
	bogl_clipminy = 0;
	bogl_clipmaxy = -1;
	bogl_clipresult = 0;
	return 0;
  }
  if (x >= bogl_xres) {
	bogl_clipminx = bogl_xres;
	bogl_clipmaxx = 65535;
	bogl_clipminy = 0;
	bogl_clipmaxy = 65535;
	bogl_clipresult = 0;
	return 0;
  }
  if (y >= bogl_yres) {
	bogl_clipminx = 0;
	bogl_clipmaxx = 65535;
	bogl_clipminy = bogl_yres;
	bogl_clipmaxy = 65535;
	bogl_clipresult = 0;
	return 0;
  }

  /* The point is within the screen area. If there are no clip
   * rectangles, then the point is plottable and the rectangle is the
   * whole screen.
   */
  count = clipcount;
  if (count <= 0) {
	bogl_clipminx = 0;
	bogl_clipmaxx = bogl_xres - 1;
	bogl_clipminy = 0;
	bogl_clipmaxy = bogl_yres - 1;
	bogl_clipresult = 1;
	return 1;
  }

  /* We need to scan the list of clip rectangles to calculate a new
   * clip cache rectangle containing this point, and the result. First
   * see if the point lies within any of the clip rectangles. If so,
   * then it is plottable and use that clip rectangle as the cache
   * rectangle.  This is not necessarily the best result, but works ok
   * and is fast.
   */
  for (rp = cliprects; count-- > 0; rp++) {
	if ((x >= rp->x) && (y >= rp->y) && (x < rp->x + rp->width)
	    && (y < rp->y + rp->height)) {
		bogl_clipminx = rp->x;
		bogl_clipminy = rp->y;
		bogl_clipmaxx = rp->x + rp->width - 1;
		bogl_clipmaxy = rp->y + rp->height - 1;
		bogl_clipresult = 1;
		return 1;
	}
  }

  /* The point is not plottable. Scan the clip rectangles again to
   * determine a rectangle containing more non-plottable points.
   * Simply pick the largest rectangle whose area doesn't contain any
   * of the same coordinates as appropriate sides of the clip
   * rectangles.  This is not necessarily the best result, but works ok
   * and is fast.
   */
  bogl_clipminx = 0;
  bogl_clipminy = 0;
  bogl_clipmaxx = 65535;
  bogl_clipmaxy = 65535;
  count = clipcount;
  for (rp = cliprects; count-- > 0; rp++) {
	if ((x < rp->x) && (rp->x <= bogl_clipmaxx)) bogl_clipmaxx = rp->x - 1;
	temp = rp->x + rp->width - 1;
	if ((x > temp) && (temp >= bogl_clipminx)) bogl_clipminx = temp + 1;
	if ((y < rp->y) && (rp->y <= bogl_clipmaxy)) bogl_clipmaxy = rp->y - 1;
	temp = rp->y + rp->height - 1;
	if ((y > temp) && (temp >= bogl_clipminy)) bogl_clipminy = temp + 1;
  }
  bogl_clipresult = 0;
  return 0;
}


/*===========================================================================*
 *				cliparea				     *
 *===========================================================================*/
int bogl_cliparea(int x1, int y1, int x2, int y2)
{
/* Check the area determined by the specified pair of points against the
 * list of clip rectangles.  The area will either be totally visible,
 * totally visible, or possibly partially visible.  This routine updates
 * the clip cache rectangle, and returns one of the following values:
 *	GR_CLIP_VISIBLE		The whole rectangle is visible
 *	GR_CLIP_INVISIBLE	The whole rectangle is invisible
 *	GR_CLIP_PARTIAL		The rectangle may be partially visible
 */
  if ((x1 < bogl_clipminx) || (x1 > bogl_clipmaxx) ||
      (y1 < bogl_clipminy) || (y1 > bogl_clipmaxy))
	bogl_clippoint(x1, y1);

  if ((x2 >= bogl_clipminx) && (x2 <= bogl_clipmaxx) &&
      (y2 >= bogl_clipminy) && (y2 <= bogl_clipmaxy)) {
	if (!bogl_clipresult)
		return GR_CLIP_INVISIBLE;
	return GR_CLIP_VISIBLE;
  }
  return GR_CLIP_PARTIAL;
}

/* END CODE */
