/* BOGL - Ben's Own Graphics Library.
   Written by Ben Pfaff <pfaffben@debian.org>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA. */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "bowl.h"

#ifndef unused
#define unused __attribute__((unused))
#endif

int 
stderrToTTY (int tty)
{
  static int fd = -1;
  char dev[10];

  /* stderr redirect only works on 2.0 kernels */
  /* FIXME: we need to figure out why it fails on 2.2 kernels.
   *        In the meantime skip it. */
#ifndef STANDALONE_TEST
  if (strncmp (real_kver, "2.0", 3))
#endif
    return 0;
  if (fd > 0)
    close (fd);
  snprintf (dev, 10, "/dev/tty%d", tty);
  if ((fd = open (dev, O_RDWR | O_NOCTTY)) < 0)
    return 1;
  if (-1 == dup2 (fd, 2))
    return 1;
  return 0;
}

void 
boxResume (void)
{
  stderrToTTY (3);
  bowl_init ();
}

void 
boxSuspend (void)
{
  bowl_done ();
  stderrToTTY (1);
}

void 
boxPopWindow (void)
{
  /* Not yet implemented. */
}

void 
boxFinished (void)
{
  bowl_done ();
  stderrToTTY (1);
}

void
initMonoScreen (void) 
{
  /* Print 24 blank lines right before erasing the screen. That way,
     the user may see the boot messages using Shift-PgUp. */
  printf("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
  bowl_init ();
  stderrToTTY(3);
}

int
pleaseWaitBox (const char *text)
{
  bowl_flush ();
  bowl_title (MSG_PLEASE_WAIT);
  bowl_new_text (text);
  bowl_layout ();

  return 0;
}

int
problemBox (const char *text, const char *title)
{
  bowl_flush ();
  bowl_title (title);
  bowl_new_text (text);
  bowl_new_button (MSG_CONTINUE, 0);
  bowl_layout ();
  bowl_run ();

  return 0;
}

int
wideMessageBox (const char *text, const char *title)
{
  bowl_flush ();
  bowl_title (title);
  bowl_new_text (text);
  bowl_new_button (MSG_CONTINUE, 0);
  bowl_layout ();
  bowl_run ();

  return 0;
}

int
perrorBox (const char *text)
{
  char buf[1024];
  snprintf (buf, 1024, "%s: %s", text, strerror (errno));
  problemBox (buf, MSG_ERROR);
  return 0;
}

int
twoButtonBox (const char *text, const char *title,
	      const char *button1, const char *button2)
{
  bowl_flush ();
  bowl_title (title);
  bowl_new_text (text);
  bowl_new_button (button1, 1);
  bowl_new_button (button2, 0);
  bowl_layout ();
  return bowl_run ();
}

int
yesNoBox (const char *text, const char *title)
{
  return twoButtonBox (text, title, MSG_YES, MSG_NO);
}

char *
inputBox (const char *text, const char *title, const char *proto)
{
  char *s;

  bowl_flush ();
  bowl_title (title);
  bowl_new_text (text);
  bowl_new_input (&s, proto);
  bowl_new_button (MSG_OK, 1);
  bowl_new_button (MSG_CANCEL, 0);
  bowl_layout ();
  bowl_default (1);
  if (bowl_run ())
    return s;
  free (s);
  return NULL;
}

int
menuBox (const char *text, const char *title,
	 struct d_choices *choices, int nchoices, int cancel)
{
  int i;

  for (i = 0; i < nchoices; i++)
    if (choices[i].tag || choices[i].string)
      choices[i].state = i;
    else
      choices[i].state = -1;
  
  bowl_flush ();
  bowl_title (title);
  bowl_new_text (text);
  bowl_new_menu ((struct bowl_menu_item *) choices, nchoices, 15);
  if (cancel)
    bowl_new_button (MSG_CANCEL, -1);
  bowl_layout ();
  return bowl_run ();
}

int 
scaleBox (const char *text, const char *title, long long value, int action)
{
  static struct widget *scale;

  switch (action)
    {
    case SCALE_CREATE:
      bowl_flush ();
      bowl_title (title);
      bowl_new_text (text);
      scale = bowl_new_scale (value);
      bowl_layout ();
      bowl_refresh ();
      break;

    case SCALE_REFRESH:
      bowl_set_scale (scale, value);
      break;

    case SCALE_DELETE:
      bowl_flush ();
      bowl_refresh ();
      break;
    }

  return 0;
}

int
check_box (const char *text, const char *title, int height unused,
	   int width unused, char **choices, char **values, int nchoices)
{
  char *result;

  result = malloc (nchoices);
  memcpy (result, *values, nchoices);

  bowl_flush ();
  bowl_title (title);
  bowl_new_text (text);
  bowl_new_checkbox (choices, result, nchoices, 15);
  bowl_new_button (MSG_OK, 1);
  bowl_new_button (MSG_CANCEL, 0);
  bowl_default (1);
  bowl_layout ();
  if (bowl_run ())
    {
      memcpy (*values, result, nchoices);
      return DLG_CANCEL;
    }
  else
    return DLG_OKAY;
}

#ifdef STANDALONE_TEST
int
main (void)
{
  initMonoScreen ();
#if 0
  twoButtonBox (MSG_NOT_1ST_DISK, MSG_CONFIRMATION, MSG_YES, MSG_NO);
#elif 0
  problemBox (MSG_SILO_PROBLEM, MSG_PROBLEM);
#elif 0
  for (;;)
    {
      pleaseWaitBox (MSG_RUNNING_LILO);
      sleep (2);
    }
#elif 0
  inputBox (MSG_LINUX_AND_SWAP, MSG_NO_SWAP_PARTITION, "/dev/sda1");
#elif 0
  {
    struct d_choices opt[30];
    int i;

    for (i = 0; i < 30; i++)
      {
	opt[i].tag = malloc (16);
	sprintf (opt[i].tag, "tag %d", i);
	opt[i].string = malloc (128);
	sprintf (opt[i].string, "item %d", i);
	opt[i].state = i;
      }

    menuBox ("This is a little text", "I'm the title", opt, 30, 1);
  }
#elif 0
  {
    char *choices[30];
    char values[30];
    char *valuesp;
    int i;

    for (i = 0; i < 30; i++)
      {
	choices[i] = malloc (128);
	sprintf (choices[i], "item %d", i);
	values[i] = i % 2 ? '*' : ' ';
      }

    valuesp = values;
    check_box ("This is a little text", "I'm the title", 10, 50,
	       choices, &valuesp, 30);
  }
#elif 0
  {
    int i;

    scaleBox ("Installing rescue floppy...", "Please wait",
	      10, SCALE_CREATE);
    for (i = 0; i <= 10; i++)
      {
	scaleBox (NULL, NULL, i, SCALE_REFRESH);
	sleep (1);
      }
  }
#elif 1
  bowl_new_text (MSG_BAD_FLOPPY MSG_LINUX_AND_SWAP);
  bowl_title (MSG_NO_SWAP_PARTITION);
  bowl_new_button (MSG_YES, 0);
  bowl_new_button (MSG_NO, 0);
  bowl_new_button (MSG_CANCEL, 0);
  bowl_new_button ("Quit", 0);
  bowl_new_button ("Help", 0);
  bowl_layout ();
  bowl_run ();
#endif

  return 0;
}
#endif
