/*
 * Copyright (c) 1991 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routines to package up drawing messages to be sent to the device driver.
 * These routines are the low level graphics driver routines, and should
 * not be called by clients.  Clients must call the GrXXXX routines instead
 * of these GdXXXX routines (or the GsXXXX routines too).
 */

#include "graph_serv.h"
#include <sys/types.h>
#include <fcntl.h>

#include "bogl/bogl.h"

extern struct bogl_font font_helvB10;

/*
 * The following definition is useful for debugging.
 * Making it nonzero disables actual writing to the graphics device,
 * thus preserving the screen and allowing you to see printfs that
 * your program is doing.
 */
#define	DEBUG	0

#define	NULL	0


int GdOpenScreen(GR_SIZE rows, GR_SIZE cols, long colors)
{
	if(bogl_init()==0)
		return GR_FALSE;
	bogl_clear(0,0, bogl_xres,bogl_yres,0);
	return GR_TRUE;
}


/*
 * Close the graphics device, first flushing any waiting messages.
 */
void GdCloseScreen(void)
{
	bogl_done();
}


static int error = 0;

/*
 * Check to see if there was a device error.
 * Clears the error if indicated.
 */
GR_BOOL GdCheckError(GR_BOOL clearflag)
{
	if (!error)
		return GR_FALSE;
	if (clearflag)
		error = GR_FALSE;
	return GR_TRUE;
}


/*
 * Flush the message buffer of any messages it may contain.
 */
void GdFlush(void)
{
}


/*
 * Terminate graphics mode, but keep the graphics device open.
 */
void GdTextMode(void)
{
	bogl_done();
}


/*
 * Return information about the screen for clients to use.
 * This does a write to the device, then a read to get the results.
 */
void GdGetScreenInfo(GR_SCREEN_INFO *sip)
{
	sip->rows = bogl_yres;
	sip->cols = bogl_xres;
	sip->xdpcm = 1;
	sip->ydpcm = 1;
	sip->maxcolor = bogl_ncols;
	sip->black = 0;
	sip->white = 1;
	sip->buttons = 0;
	sip->modifiers = 0;
	sip->fonts = 4;
}


/*
 * Read information about a specified font.
 */
void GdGetFontInfo(GR_FONT font, GR_FONT_INFO *fip)
{
	/* FIXME - keep a bogl font array to index */
	fip->height = 16;
	fip->maxwidth = 16;
	fip->baseline = 0;
	fip->fixed = GR_FALSE;
	memcpy(fip->widths,font_helvB10.width,256);
}


/*
 * Set an array of clip rectangles for future drawing actions.
 * Each pixel will be drawn only if lies in one or more of the specified
 * clip rectangles.  As a special case, specifying no rectangles implies
 * clipping is for the complete screen.
 */

void GdSetClipRects(GR_COUNT count,  GR_RECT *cliprects)
{
	bogl_clipinit(count, (struct bogl_cliprect *)cliprects);
}


/*
 * Set the cursor size and bitmaps.
 */
void GdSetCursor(width, height, foreground, background, fgbitmap, bgbitmap)
	GR_SIZE		width;		/* width of cursor */
	GR_SIZE		height;		/* height of cursor */
	GR_COLOR	foreground;	/* foreground color of cursor */
	GR_COLOR	background;	/* background color of cursor */
	GR_BITMAP	*fgbitmap;	/* foreground bitmap */
	GR_BITMAP	*bgbitmap;	/* background bitmap */
{
}


/*
 * Set the cursor position.
 */
void GdMoveCursor(x, y)
	GR_COORD	x;		/* new x position of cursor */
	GR_COORD	y;		/* new y position of cursor */
{
}


/*
 * Set whether or not the background is used for drawing pixmaps and text.
 */
void
GdSetUseBackground(flag)
	GR_BOOL		flag;		/* TRUE if background is drawn */
{
}


static GR_COLOR foreground, background;
static GR_MODE mode;

/*
 * Set the drawing mode for future calls.
 */
void
GdSetMode(m)
	GR_MODE		m;		/* drawing mode */
{
	mode=m;
}


/*
 * Set the foreground color for future calls.
 */
void
GdSetForeground(fg)
	GR_COLOR	fg;		/* foreground color */
{
	foreground = fg;
}


/*
 * Set the background color for future calls.
 */
void
GdSetBackground(bg)
	GR_COLOR	bg;		/* background color */
{
	background = bg;
}


/*
 * Draw a line using the current clipping region and foreground color.
 */
void
GdLine(x1, y1, x2, y2)
	GR_COORD	x1;
	GR_COORD	y1;
	GR_COORD	x2;
	GR_COORD	y2;
{
	if(x1==x2)
		bogl_vline(x1,y1,y2, foreground);
	else if(y1==y2)
		bogl_hline(x1,x2,y1, foreground);
	else
		bogl_line(x1,y1,x2,y2,foreground);
}


/*
 * Draw a rectangle using the current clipping region and foreground color.
 * This just draws in the outline of the rectangle.
 */
void GdRect(x, y, width, height)
	GR_COORD	x;
	GR_COORD	y;
	GR_SIZE		width;
	GR_SIZE		height;
{
	if(!height || !width)
		return;
	height--;
	width--;
	bogl_hline(x, x+width, y, foreground);
	bogl_hline(x, x+width, y+height, foreground);
	bogl_vline(x, y, y+height, foreground);
	bogl_vline(x+width, y, y+height, foreground);
}


/*
 * Fill a rectangle using the current clipping region and foreground color.
 * This draws the boarder plus all of its interior.
 */
void
GdFillRect(x, y, width, height)
	GR_COORD	x;
	GR_COORD	y;
	GR_SIZE		width;
	GR_SIZE		height;
{
	bogl_rectangle(x,y, x+width-1, y+height-1, foreground);
}


/*
 * Draw an ellipse using the current clipping region and foreground color.
 * This just draws in the outline of the ellipse.
 */
void
GdEllipse(x, y, rx, ry)
	GR_COORD	x;
	GR_COORD	y;
	GR_SIZE		rx;
	GR_SIZE		ry;
{
	bogl_drawellipse(x,y,rx,ry, foreground);
}


/*
 * Fill an ellipse using the current clipping region and foreground color.
 */
void
GdFillEllipse(x, y, rx, ry)
	GR_COORD	x;
	GR_COORD	y;
	GR_SIZE		rx;
	GR_SIZE		ry;
{
	bogl_fillellipse(x,y,rx,ry, foreground);
}


/*
 * Copy a rectangular area from one location on the screen to another.
 * This bypasses clipping.
 */
void
GdCopyArea(srcx, srcy, width, height, destx, desty)
	GR_COORD	srcx;
	GR_COORD	srcy;
	GR_SIZE		width;
	GR_SIZE		height;
	GR_COORD	destx;
	GR_COORD	desty;
{
}


/*
 * Draw a rectangular area using the current clipping region and the
 * specified bit map.  This differs from rectangle drawing in that the
 * rectangle is drawn using the foreground color and possibly the background
 * color as determined by the bit map.  Each row of bits is aligned to the
 * next bitmap word boundary (so there is padding at the end of the row).
 * The background bit values are only written if the usebackground flag
 * is set in the GC.
 */
void
GdBitmap(x, y, width, height, bitmaptable)
	GR_COORD	x;
	GR_COORD	y;
	GR_SIZE		width;
	GR_SIZE		height;
	GR_BITMAP	*bitmaptable;
{
}


/*
 * Draw a rectangular area using the current clipping region.
 * This differs from rectangle drawing in that the color values for
 * each pixel in the rectangle is specified.  The color values are
 * restricted to 8 bit values.  The color table is indexed row by row.
 */
void
GdArea8(x, y, width, height, colortable)
	GR_COORD	x;
	GR_COORD	y;
	GR_SIZE		width;
	GR_SIZE		height;
	GR_COLOR8	*colortable;
{
}


/*
 * Read a rectangular area of the screen.  The color values are restricted
 * to 8 bit values.  The color table is indexed row by row.
 */
void
GdReadArea8(x, y, width, height, colortable)
	GR_COORD	x;
	GR_COORD	y;
	GR_SIZE		width;
	GR_SIZE		height;
	GR_COLOR8	*colortable;
{
}


/*
 * Draw a point using the current clipping region and foreground color.
 */
void
GdPoint(x, y)
	GR_COORD	x;
	GR_COORD	y;
{
	bogl_pixel(x, y, foreground);
}


/*
 * Draw a polygon from the specified list of points.  The polygon is
 * only complete if the first point is repeated at the end.
 */
void
GdPoly(count, pointtable)
	GR_COUNT	count;
	GR_POINT	*pointtable;
{
}


/*
 * Draw a filled polygon from the specified list of points.  The last
 * point may be a duplicate of the first point, but this is not required.
 */
void GdFillPoly(count, pointtable)
	GR_COUNT	count;
	GR_POINT	*pointtable;
{
}


/*
 * Draw a text string starting at the specified coordinates.
 */
void
GdText(x, y, str, count)
	GR_COORD	x;
	GR_COORD	y;
	GR_CHAR		*str;
	GR_COUNT	count;
{
	bogl_text(x, y, str, count, foreground, background,  &font_helvB10);
}

/* END CODE */
